"""
Set of util functions and classes
"""


class literal(str): pass


def literal_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style="|")


class Serializable(object):

    @staticmethod
    def serialize(obj):
        if isinstance(obj, str) or isinstance(obj, int) or isinstance(obj, float) or isinstance(obj, bool):
            return obj
        if isinstance(obj, list):
            return [Serializable.serialize(x) for x in obj]
        if isinstance(obj, dict):
            return {k: Serializable.serialize(v) for k, v in obj.items() if v is not None}
        if callable(getattr(obj, "get_dict", None)):
            return obj.get_dict()
        raise Exception("Can't serialize '{unsupported_type}'".format(unsupported_type=type(obj)))

    @staticmethod
    def should_exclude_object(value):
        return value is None or ((isinstance(value, list) or isinstance(value, dict)) and len(value) == 0)

    def get_dict(self):
        d = {}
        ignore_attributes = set()
        if hasattr(self, "ignore_attributes") and self.ignore_attributes:
            ignore_attributes = set(self.ignore_attributes)

        for k, v in vars(self).items():
            if self.should_exclude_object(v) or k.startswith("_") or k in ignore_attributes or k == "ignore_attributes":
                continue
            s = self.serialize(v)
            if self.should_exclude_object(s):
                continue
            d[k] = s
        return d
        # return {k: self.serialize(v) for k, v in vars(self).items() if v is not None}



#!/usr/bin/env python

'''
Unit tests for cwlgen library
'''

#  Import  ------------------------------

# General libraries
import os
import filecmp
import unittest
from tempfile import NamedTemporaryFile

# External libraries
import cwlgen

#  Class(es)  ------------------------------

class TWorkflow:
# class TestWorkflow(unittest.TestCase):

    def capture_tempfile(self, func):
        with NamedTemporaryFile() as f:
            func(f.name)
            return f.read()

    def test_generates_workflow_two_steps(self):

        w = cwlgen.Workflow("test_generates_workflow_two_steps")
        tool = cwlgen.parse_cwl("test/import_cwl.cwl")

        f = cwlgen.File("input_file")

        o1 = w.add('step-a', tool, {"INPUT1" : f})
        o2 = w.add('step-b', tool, {"INPUT1" : o1['OUTPUT1']})
        o2['OUTPUT1'].store()
        generated = self.capture_tempfile(w.export)
        expected = b"""#!/usr/bin/env cwl-runner

class: Workflow
cwlVersion: v1.0
inputs:
  INPUT1: {id: INPUT1, type: File}
outputs:
  step-b_OUTPUT1: {id: step-b_OUTPUT1, outputSource: step-b/OUTPUT1, type: File}
steps:
  step-a:
    id: step-a
    in: {INPUT1: INPUT1}
    out: [OUTPUT1]
    run: test/import_cwl.cwl
  step-b:
    id: step-b
    in: {INPUT1: step-a/OUTPUT1}
    out: [OUTPUT1]
    run: test/import_cwl.cwl
"""
        self.assertEqual(expected, generated)

    def test_generates_workflow_int_inputs(self):

        w = cwlgen.Workflow("test_generates_workflow_int_inputs")
        tool = cwlgen.parse_cwl("test/int_tool.cwl")

        i = cwlgen.workflow.InputParameter('INTEGER', param_type='int')
        o1 = w.add('step', tool, {"INTEGER": i})
        o1['OUTPUT1'].store()

        expected = b"""#!/usr/bin/env cwl-runner

class: Workflow
cwlVersion: v1.0
inputs:
  INTEGER: {id: INTEGER, type: int}
outputs:
  step_OUTPUT1: {id: step_OUTPUT1, outputSource: step/OUTPUT1, type: File}
steps:
  step:
    id: step
    in: {INTEGER: INTEGER}
    out: [OUTPUT1]
    run: test/int_tool.cwl
"""
        generated = self.capture_tempfile(w.export)
        self.assertEqual(expected, generated)

    def test_add_requirements(self):
        w = cwlgen.Workflow("test_add_requirements")
        req = cwlgen.InlineJavascriptReq()
        w.requirements.append(req)
        generated = self.capture_tempfile(w.export)
        expected = b"""#!/usr/bin/env cwl-runner

class: Workflow
cwlVersion: v1.0
inputs: {}
outputs: {}
requirements:
  InlineJavascriptRequirement: {}
"""
        self.assertEqual(expected, generated)




#  Import  ------------------------------

# General libraries
import logging

# External libraries
import ruamel.yaml
import six

from .version import __version__

# Internal libraries

from .utils import literal, literal_presenter, Serializable
from .common import Parameter, CWL_SHEBANG


# Logging setup

logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)

SCATTER_METHODS = ["dotproduct", "nested_crossproduct", "flat_crossproduct"]
LINK_MERGE_METHODS = ["merge_nested", "merge_flattened"]


#  Function(s)  ------------------------------

def parse_scatter_method(scatter_method, required=False):
    if scatter_method is None and not required:
        return None
    elif scatter_method not in SCATTER_METHODS:
        if required:
            raise Exception("The scatter method '{method}' is not a valid ScatterMethod and requires one of: {expected}"
                            .format(method=scatter_method, expected=" ,".join(SCATTER_METHODS)))
        elif scatter_method is not None:
            _LOGGER.info("The scatter method '{method}' is not a valid ScatterMethod, expected one of: {expected}"
                         .format(method=scatter_method, expected=" ,".join(SCATTER_METHODS)))
            return None
    return scatter_method


def parse_link_merge_method(link_merge, required=False):
    if link_merge is None and not required:
        return None
    elif link_merge not in LINK_MERGE_METHODS:
        if required:
            raise Exception("The link merge method '{method}' is not a valid LinkMergeMethod and requires one of:"
                            " {expected}. ".format(method=link_merge, expected=" ,".join(LINK_MERGE_METHODS)))
        elif link_merge is not None:
            _LOGGER.info("The link merge method '{method}' is not a valid LinkMergeMethod, expected one of:"
                            " {expected}. This value will be null which CWL defaults to 'merge_nested'"
                            .format(method=link_merge, expected=" ,".join(LINK_MERGE_METHODS)))
            return None
    return link_merge


#  Class(es)  ------------------------------

class Workflow(Serializable):
    """
    A workflow describes a set of steps and the dependencies between those steps.
    When a step produces output that will be consumed by a second step,
    the first step is a dependency of the second step.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#Workflow
    """
    __CLASS__ = 'Workflow'

    def __init__(self, workflow_id=None, label=None, doc=None, cwl_version='v1.0'):
        """
        :param workflow_id: The unique identifier for this process object.
        :type workflow_id: STRING
        :param label: A short, human-readable label of this process object.
        :type label: STRING
        :param doc: A long, human-readable description of this process object.
        :type doc: STRING
        :param cwl_version: CWL document version. Always required at the document root. Default: 'v1.0'
        :type cwl_version: CWLVersion
        """
        self.id = workflow_id
        self.label = label
        self.doc = doc
        self.cwlVersion = cwl_version

        self.inputs = []            # list[InputParameter]
        self.outputs = []           # list[WorkflowOutputParameter]
        self.steps = []             # list[WorkflowStep]
        self.requirements = []
        self.hints = None
        self._path = None

        self.ignore_attributes = ["inputs", "outputs"]

    def get_dict(self):
        cwl_workflow = super(Workflow, self).get_dict()

        cwl_workflow['class'] = self.__CLASS__

        cwl_workflow['inputs'] = {}
        cwl_workflow['outputs'] = {}

        # steps, inputs, outputs are required properties, so it should fail if we can't place it
        cwl_workflow['steps'] = {step.id: step.get_dict() for step in self.steps}
        cwl_workflow['inputs'] = {i.id: i.get_dict() for i in self.inputs}
        cwl_workflow['outputs'] = {o.id: o.get_dict() for o in self.outputs}

        if self.requirements:
            cwl_workflow['requirements'] = {r.get_class(): r.get_dict() for r in self.requirements}

        return cwl_workflow

    def export(self, outfile=None):
        """
        Export the workflow in CWL either on STDOUT or in outfile.
        """
        # First add representer (see .utils.py) for multiline writting
        ruamel.yaml.add_representer(literal, literal_presenter)

        cwl_workflow = self.get_dict()

        # Write CWL file in YAML
        if outfile is None:
            six.print_(CWL_SHEBANG, "\n", sep='')
            six.print_(ruamel.yaml.dump(cwl_workflow))
        else:
            out_write = open(outfile, 'w')
            out_write.write(CWL_SHEBANG + '\n\n')
            out_write.write(ruamel.yaml.dump(cwl_workflow))
            out_write.close()

    # def add(self, step_id, tool, params):
    #     return StepRun(self, step_id, tool, params)


class InputParameter(Parameter):
    """
    Documentation: https://www.commonwl.org/v1.0/Workflow.html#InputParameter
    """
    def __init__(self, param_id, label=None, secondary_files=None, param_format=None,
                 streamable=None, doc=None, input_binding=None, default=None, param_type=None):
        """
        :param param_id: unique identifier for this parameter
        :type param_id: STRING
        :param label: short, human-readable label
        :type label: STRING
        :param secondary_files: If type is a file, describes files that must be
                                included alongside the primary file(s)
        :type secondary_files: STRING
        :param param_format: If type is a file, uri to ontology of the format or exact format
        :type param_format: STRING
        :param streamable: If type is a file, true indicates that the file is read or written
                           sequentially without seeking
        :type streamable: BOOLEAN
        :param doc: documentation
        :type doc: STRING
        :param param_type: type of data assigned to the parameter
        :type param_type: STRING corresponding to CWLType
        :param input_binding:
        :type input_binding: CommandLineBinding

        """
        Parameter.__init__(self, param_id=param_id, label=label,
                           secondary_files=secondary_files, param_format=param_format,
                           streamable=streamable, doc=doc, param_type=param_type, requires_type=False)
        self.inputBinding = input_binding
        self.default = default


class WorkflowStep(Serializable):
    """
    A workflow step is an executable element of a workflow. It specifies the underlying process implementation
    (such as CommandLineTool or another Workflow) in the run field and connects the input and output parameters
    of the underlying process to workflow parameters.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#WorkflowStep
    """
    def __init__(self, step_id, run, label=None, doc=None, scatter=None, scatter_method=None):
        """
        :param step_id: The unique identifier for this workflow step.
        :type step_id: STRING
        :param run: Specifies the process to run.
        :type run: STRING | CommandLineTool | ExpressionTool | Workflow
        :param label: A short, human-readable label of this process object.
        :type label: STRING
        :param doc: A long, human-readable description of this process object.
        :type doc: STRING | list[STRING]
        :param scatter: Field to scatter on, see: https://www.commonwl.org/v1.0/Workflow.html#WorkflowStep
        :type scatter: STRING | list[STRING]
        :param scatter_method: Required if scatter is an array of more than one element.
        :type scatter_method: STRING | list[STRING] in [dotproduct, nested_crossproduct, flat_crossproduct]
        """
        self.id = step_id
        self.run = run
        self.label = label
        self.doc = doc
        self.scatter = scatter
        self.scatterMethod = parse_scatter_method(scatter_method)

        # in is a reserved keywork
        self.inputs = []
        self.out = []
        self.requirements = []
        self.hints = []

        self.ignore_attributes = ["id", "inputs"]

    def get_dict(self):
        d = super(WorkflowStep, self).get_dict()
        d['in'] = {i.id: self.serialize(i) for i in self.inputs}
        return d


class WorkflowStepInput(Serializable):
    """
    The input of a workflow step connects an upstream parameter (from the workflow inputs, or the outputs of
    other workflows steps) with the input parameters of the underlying step.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#WorkflowStepInput
    """
    def __init__(self, input_id, source=None, link_merge=None, default=None, value_from=None):
        """
        :param input_id: A unique identifier for this workflow input parameter.
        :type input_id: STRING
        :param source: Specifies one or more workflow parameters that will provide input to the underlying step parameter.
        :type source: STRING | list[STRING]
        :param link_merge: The method to use to merge multiple inbound links into a single array.
                           If not specified, the default method is "merge_nested".
        :type link_merge: LinkMergeMethod
        :param default: The default value for this parameter to use if either there is no source field,
                        or the value produced by the source is null
        :type default: Any | DictRepresentible
        :param value_from: If valueFrom is a constant string value, use this as the value for this input parameter.
                           If valueFrom is a parameter reference or expression,
                           it must be evaluated to yield the actual value to be assiged to the input field.
        :type value_from: STRING
        """

        self.id = input_id
        self.source = source
        self.linkMerge = parse_link_merge_method(link_merge)
        self.default = default
        self.valueFrom = value_from


class WorkflowStepOutput(Serializable):
    """
    Associate an output parameter of the underlying process with a workflow parameter.
    The workflow parameter (given in the id field) be may be used as a source to connect with
    input parameters of other workflow steps, or with an output parameter of the process.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#WorkflowStepOutput
    """
    def __init__(self, output_id):
        """
        :param output_id: A unique identifier for this workflow output parameter. This is the identifier to use in
        the source field of WorkflowStepInput to connect the output value to downstream parameters.
        :type output_id: STRING
        """
        self.id = output_id

    def get_dict(self):
        return self.id


class WorkflowOutputParameter(Parameter):
    """
    Describe an output parameter of a workflow. The parameter must be connected to one or more parameters
    defined in the workflow that will provide the value of the output parameter.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#WorkflowOutputParameter
    """
    def __init__(self, param_id, output_source=None, label=None, secondary_files=None, param_format=None,
                 streamable=None, doc=None, param_type=None, output_binding=None, linkMerge=None):
        """
        Documentation: https://www.commonwl.org/v1.0/Workflow.html#WorkflowOutputParameter
        :param param_id: The unique identifier for this parameter object.
        :type param_id: STRING
        :param output_source: Specifies one or more workflow parameters that supply the value of to the output parameter
        :type output_source: STRING | list[STRING]
        :param label: A short, human-readable label of this object.
        :type label: STRING
        :param secondary_files: Provides a pattern or expression specifying files or directories that must be
                                included alongside the primary file.
        :type secondary_files: STRING \ list[STRING]
        :param param_format: This is the file format that will be assigned to the output paramete
        :type param_format: STRING
        :param streamable: A value of true indicates that the file is read or written sequentially without seeking
        :type streamable: BOOLEAN
        :param doc: A documentation string for this type, or an array of strings which should be concatenated.
        :type doc: STRING | list[STRING]
        :param param_type: Specify valid types of data that may be assigned to this parameter.
        :type param_type: CWLType | OutputRecordSchema | OutputEnumSchema | OutputArraySchema | string | Array<type>
        :param output_binding: Describes how to handle the outputs of a process.
        :type output_binding: CommandOutputBinding
        :param linkMerge:
        :type linkMerge: STRING
        """
        Parameter.__init__(self, param_id=param_id, label=label,
                           secondary_files=secondary_files, param_format=param_format,
                           streamable=streamable, doc=doc, param_type=param_type, requires_type=False)
        self.outputSource = output_source
        self.outputBinding = output_binding     # CommandOutputBinding
        self.linkMerge = linkMerge


############################
# Workflow construction classes

# class File:
#     """
#     An abstract file reference used for generating workflows
#     """
#     def __init__(self, path):
#         self.path = path
#
#
# class Variable:
#     """
#     An output variable from a workflow step
#     """
#     def __init__(self, workflow, step, name):
#         self.step = step
#         self.name = name
#         self.workflow = workflow
#
#     def path(self):
#         return "%s/%s" % (self.step, self.name)
#
#     def store(self):
#         self.workflow.outputs.append(
#             WorkflowOutputParameter(self.path().replace("/", "_"),
#                                     outputSource=self.path(),
#                                     param_type="File"))
#         return


# class StepRun:
#     """
#     Result of adding a step into a workflow
#     """
#     def __init__(self, workflow, id, tool, params):
#         self.tool = tool
#         self.workflow = workflow
#         self.id = id
#
#         step = WorkflowStep(id=id, run=tool._path)
#         workflow.steps.append(step)
#
#         for i, j in params.items():
#             if isinstance(j, six.string_types):
#                 step.inputs.append(WorkflowStepInput(i, default=j))
#             elif isinstance(j, Variable):
#                 step.inputs.append(WorkflowStepInput(i, src=j.path()))
#             elif isinstance(j, InputParameter):
#                 self.workflow.inputs.append(j),
#                 step.inputs.append(WorkflowStepInput(j.id, src=j.id))
#             elif isinstance(j, File):
#                 # This is just used as a stub, the 'path' inside the file doesn't do anything
#                 self.workflow.inputs.append(InputParameter(i, param_type="File"))
#                 step.inputs.append(WorkflowStepInput(i, src=i))
#         for o in tool.outputs:
#             step.outputs.append(o.id)
#
#     def store_all(self):
#         for i in self.tool.outputs:
#             Variable(self.workflow, self.id, i.id).store()
#
#     def __getitem__(self, key):
#         for i in self.tool.outputs:
#             if i.id == key:
#                 return Variable(self.workflow, self.id, key)
#         raise KeyError



import six

from .common import parse_type, get_type_dict
from .utils import Serializable


class Requirement(Serializable):
    '''
    Requirement that must be met in order to execute the process.
    '''

    def __init__(self, req_class):
        '''
        :param req_class: requirement class
        :type req_class: STRING
        '''
        # class is protected keyword
        self._req_class = req_class

    def get_class(self):
        return self._req_class


class InlineJavascriptReq(Requirement):
    """
    Indicates that the workflow platform must support inline Javascript expressions.
    If this requirement is not present, the workflow platform must not perform expression interpolatation.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#InlineJavascriptRequirement
    """

    def __init__(self, expression_lib=None):
        '''
        :param expression_lib: List of Strings
        :type expression_lib: list[STRING]
        '''
        Requirement.__init__(self, 'InlineJavascriptRequirement')
        self.expressionLib = [expression_lib] if isinstance(expression_lib, six.string_types) else expression_lib


class SchemaDefRequirement(Requirement):
    """
    This field consists of an array of type definitions which must be used when interpreting the inputs and
    outputs fields. When a type field contain a IRI, the implementation must check if the type is defined
    in schemaDefs and use that definition. If the type is not found in schemaDefs, it is an error.
    The entries in schemaDefs must be processed in the order listed such that later schema definitions
    may refer to earlier schema definitions.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#SchemaDefRequirement
    """

    def __init__(self, types):
        """
        :param types: The list of type definitions.
        :type types: list[InputRecordSchema | InputEnumSchema | InputArraySchema]
        """
        Requirement.__init__(self, "SchemaDefRequirement")
        self.types = types

    class InputRecordSchema(Serializable):
        """
        Documentation: https://www.commonwl.org/v1.0/Workflow.html#InputRecordSchema
        """
        def __init__(self, label=None, name=None):
            """
            :param fields: Defines the fields of the record.
            :type fields: array<InputRecordField>
            :param label: A short, human-readable label of this object.
            :param name: NF (Name of the InputRecord)
            """
            self.fields = []
            self.label = label
            self.name = name
            self.type = "record"

        class InputRecordField(Serializable):
            """
            Documentation: https://www.commonwl.org/v1.0/Workflow.html#InputRecordField
            """
            def __init__(self, name, input_type, doc=None, input_binding=None, label=None):
                """
                :param name:
                :param input_type:
                :type input_type: CWLType | InputRecordSchema | InputEnumSchema | InputArraySchema | string |
                            array<CWLType | InputRecordSchema | InputEnumSchema | InputArraySchema | string>
                :param doc: A documentation string for this field
                :param input_binding:
                :type input_binding: CommandLineBinding
                :param label:
                """
                self.name = name
                self.type = parse_type(input_type, requires_type=True)
                self.doc = doc
                self.inputBinding = input_binding
                self.label = label

            def get_dict(self):
                d = super(self, self).get_dict()
                d["type"] = get_type_dict(self.type)
                return d

    class InputEnumSchema(Serializable):
        """
        Documentation: https://www.commonwl.org/v1.0/Workflow.html#InputEnumSchema
        """
        def __init__(self, symbols, label=None, name=None, input_binding=None):
            """
            :param symbols: Defines the set of valid symbols.
            :type symbols: list[STRING]
            :param label: A short, human-readable label of this object.
            :type label: STRING
            :param name:
            :type name: STRING
            :param input_binding:
            :type input_binding: CommandLineBinding
            """
            self.type = "enum"
            self.symbols = symbols
            self.label = label
            self.name = name
            self.inputBinding = input_binding

    class InputArraySchema(Serializable):
        """
        Documentation: https://www.commonwl.org/v1.0/Workflow.html#InputArraySchema
        """

        def __init__(self, items, label=None, input_binding=None):
            """
            :param items: Defines the type of the array elements.
            :type items: CWLType | InputRecordSchema | InputEnumSchema | InputArraySchema | string |
                    array<CWLType | InputRecordSchema | InputEnumSchema | InputArraySchema | string>
            :param label: A short, human-readable label of this object.
            :type label: STRING
            :param input_binding:
            :type input_binding: CommandLineBinding
            """
            self.type = "array"
            self.items = items
            self.label = label
            self.inputBinding = input_binding


class SoftwareRequirement(Requirement):
    """
    A list of software packages that should be configured in the environment of the defined process.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#SoftwareRequirement
    """
    def __init__(self):
        Requirement.__init__(self, "SoftwareRequirement")
        self.packages = []      # list[SoftwarePackage]

    class SoftwarePackage(Serializable):
        """
        Documentation: https://www.commonwl.org/v1.0/Workflow.html#SoftwarePackage
        """
        def __init__(self, package, version=None, specs=None):
            """
            :param package: The name of the software to be made available. If the name is common, inconsistent,
                            or otherwise ambiguous it should be combined with one or more identifiers in the specs field
            :param version: The (optional) versions of the software that are known to be compatible.
            :param specs: One or more IRIs identifying resources for installing or enabling the software in 'package'
            """
            self.package = package
            self.version = version
            self.specs = specs


class InitialWorkDirRequirement(Requirement):
    """
    Define a list of files and subdirectories that must be created by the workflow
    platform in the designated output directory prior to executing the command line tool.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#InitialWorkDirRequirement
    """
    def __init__(self, listing):
        """
        :param listing: The list of files or subdirectories that must be placed in the
                        designated output directory prior to executing the command line tool.
        :type listing: array<File | Directory | Dirent | string | Expression> | string | Expression
        """
        Requirement.__init__(self, "InitialWorkDirRequirement")
        self.listing = listing

    def get_dict(self):
        base = super(InitialWorkDirRequirement, self).get_dict()

        if isinstance(self.listing, str):
            base["listing"] = self.listing
        elif isinstance(self.listing, list):
            if len(self.listing) == 0:
                raise Exception("InitialWorkDirRequirement.listing must have at least one element")
            base["listing"] = [r if isinstance(r, str) else r.get_dict() for r in self.listing]
        else:
            raise Exception("Couldn't recognise type of '{list_type}', expected: array<File | Directory | Dirent | "
                            "string| Expression> | string | Expression".format(list_type=type(self.listing)))

        return base

    class Dirent(Serializable):
        """
        Define a file or subdirectory that must be placed in the designated output directory
        prior to executing the command line tool. May be the result of executing an expression,
        such as building a configuration file from a template.

        Documentation: https://www.commonwl.org/v1.0/Workflow.html#Dirent
        """
        def __init__(self, entry, entryname=None, writable=None):
            self.entry = entry
            self.entryname = entryname
            self.writable = writable


class SubworkflowFeatureRequirement(Requirement):
    """
    Indicates that the workflow platform must support nested workflows in the run field of WorkflowStep.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#SubworkflowFeatureRequirement
    """

    def __init__(self):
        Requirement.__init__(self, 'SubworkflowFeatureRequirement')


class ScatterFeatureRequirement(Requirement):
    """
    Indicates that the workflow platform must support the scatter and scatterMethod fields of WorkflowStep.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#ScatterFeatureRequirement
    """

    def __init__(self):
        Requirement.__init__(self, 'ScatterFeatureRequirement')


class MultipleInputFeatureRequirement(Requirement):
    """
    Indicates that the workflow platform must support multiple
    inbound data links listed in the source field of WorkflowStepInput.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#MultipleInputFeatureRequirement
    """

    def __init__(self):
        Requirement.__init__(self, 'MultipleInputFeatureRequirement')


class StepInputExpressionRequirement(Requirement):
    """
    Indicate that the workflow platform must support the valueFrom field of WorkflowStepInput.

    Documentation: https://www.commonwl.org/v1.0/Workflow.html#StepInputExpressionRequirement
    """

    def __init__(self):
        Requirement.__init__(self, 'StepInputExpressionRequirement')


class DockerRequirement(Requirement):
    """
    Indicates that a workflow component should be run in a Docker container,
    and specifies how to fetch or build the image.

    Documentation: https://www.commonwl.org/v1.0/CommandLineTool.html#DockerRequirement
    """

    def __init__(self, docker_pull=None, docker_load=None, docker_file=None,
                 docker_import=None, docker_image_id=None, docker_output_dir=None):
        """
        :param docker_pull: image to retrive with docker pull
        :type docker_pull: STRING
        :param docker_load: HTTP URL from which to download Docker image
        :type docker_load: STRING
        :param docker_file: supply the contents of a Dockerfile
        :type docker_file: STRING
        :param docker_import: HTTP URL to download and gunzip a Docker images
        :type docker_import: STRING
        :param docker_image_id: Image id for docker run
        :type docker_image_id: STRING
        :param docker_output_dir: designated output dir inside the Docker container
        :type docker_output_dir: STRING
        """
        Requirement.__init__(self, 'DockerRequirement')
        self.dockerPull = docker_pull
        self.dockerLoad = docker_load
        self.dockerFile = docker_file
        self.dockerImport = docker_import
        self.dockerImageId = docker_image_id
        self.dockerOutputDir = docker_output_dir


class EnvVarRequirement(Requirement):
    """
    Define a list of environment variables which will be set in the execution environment of the tool.
    See EnvironmentDef for details.

    Documentation: https://www.commonwl.org/v1.0/CommandLineTool.html#EnvVarRequirement
    """
    def __init__(self, env_def):
        """
        :param env_def: The list of environment variables.
        :type env_def: list[EnvironmentDef]
        """
        Requirement.__init__(self, 'EnvVarRequirement')
        self.envDef = env_def

    class EnvironmentDef(Serializable):
        """
        Define an environment variable that will be set in the runtime environment
        by the workflow platform when executing the command line tool.
        May be the result of executing an expression, such as getting a parameter from input.

        Documentation: https://www.commonwl.org/v1.0/CommandLineTool.html#EnvironmentDef
        """
        def __init__(self, env_name, env_value):
            """
            :param env_name: The environment variable name
            :type env_name: STRING
            :param env_value: The environment variable value
            :type env_value: STRING
            """
            self.envName = env_name
            self.envValue = env_value


class ShellCommandRequirement(Requirement):
    """
    Modify the behavior of CommandLineTool to generate a single string containing a shell command line.

    Documentation: https://www.commonwl.org/v1.0/CommandLineTool.html#ShellCommandRequirement
    """

    def __init__(self):
        Requirement.__init__(self, 'ShellCommandRequirement')


class ResourceRequirement(Requirement):
    """
    Specify basic hardware resource requirements.

    Documentation: https://www.commonwl.org/v1.0/CommandLineTool.html#ResourceRequirement
    """

    def __init__(self, cores_min=None, cores_max=None, ram_min=None, ram_max=None, tmpdir_min=None, tmpdir_max=None,
                 outdir_min=None, outdir_max=None):
        """
        :param cores_min: Minimum reserved number of CPU cores
        :type cores_min: string | float | None
        :param cores_max: Maximum reserved number of CPU cores
        :type cores_max: string | float | None
        :param ram_min: Minimum reserved RAM in mebibytes (2**20)
        :type ram_min: string | float | None
        :param ram_max: Maximum reserved RAM in mebibytes (2**20)
        :type ram_max: string | float | None
        :param tmpdir_min: Minimum reserved filesystem based storage for the designated temporary directory, in mebibytes (2**20)
        :type tmpdir_min: string | float | None
        :param tmpdir_max: Maximum reserved filesystem based storage for the designated temporary directory, in mebibytes (2**20)
        :type tmpdir_max: string | float | None
        :param outdir_min: Minimum reserved filesystem based storage for the designated output directory, in mebibytes (2**20)
        :type outdir_min: string | float | None
        :param outdir_max: Maximum reserved filesystem based storage for the designated output directory, in mebibytes (2**20)
        :type outdir_max: string | float | None
        """
        Requirement.__init__(self, 'ResourceRequirement')

        self.coresMin = cores_min
        self.coresMax = cores_max
        self.ramMin = ram_min
        self.ramMax = ram_max
        self.tmpdirMin = tmpdir_min
        self.tmpdirMax = tmpdir_max
        self.outdirMin = outdir_min
        self.outdirMax = outdir_max



#!/usr/bin/env python

'''
Unit tests for import feature of cwlgen library
'''

#  Import  ------------------------------ 

# General libraries
import os
import filecmp
import unittest

# External libraries
from cwlgen.import_cwl import CWLToolParser

#  Class(es)  ------------------------------ 

class TestImport(unittest.TestCase):

    def setUp(self):
        ctp = CWLToolParser()
        self.tool = ctp.import_cwl('test/import_cwl.cwl')


class TestImportCWL(TestImport):

    def test_load_id(self):
        self.assertEqual(self.tool.id, 'tool_id')

    def test_load_baseCommand(self):
        self.assertEqual(self.tool.baseCommand, 'share')

    def test_load_label(self):
        self.assertEqual(self.tool.label, 'a_label')

    def test_load_doc(self):
        self.assertEqual(self.tool.doc, 'super_doc')

    def test_load_cwlVersion(self):
        self.assertEqual(self.tool.cwlVersion, 'v1.0')

    def test_load_stdin(self):
        self.assertEqual(self.tool.stdin, 'in')

    def test_load_stderr(self):
        self.assertEqual(self.tool.stderr, 'err')

    def test_load_stdout(self):
        self.assertEqual(self.tool.stdout, 'out')


class TestInputsParser(TestImport):

    def test_init_input(self):
        self.assertEqual(self.tool.inputs[0].id, 'INPUT1')

    def test_load_label(self):
        self.assertEqual(self.tool.inputs[0].label, 'label_in')

    def test_load_secondaryFiles(self):
        self.assertEqual(self.tool.inputs[0].secondaryFile, 'sec_file_in')

    def test_load_format(self):
        self.assertEqual(self.tool.inputs[0].format, 'format_1930')

    def test_load_streamable(self):
        self.assertTrue(self.tool.inputs[0].streamable)

    def test_load_doc(self):
        self.assertEqual(self.tool.inputs[0].doc, 'documentation_in')

    def test_load_default(self):
        self.assertEqual(self.tool.inputs[0].default, 'def_in')

    def test_load_type(self):
        self.assertEqual(self.tool.inputs[0].type, 'File')


class TestInputBindingParser(TestImport):

    def test_load_loadContents(self):
        self.assertTrue(self.tool.inputs[0].inputBinding.loadContents)

    def test_load_position(self):
        self.assertEqual(self.tool.inputs[0].inputBinding.position, 0)

    def test_load_prefix(self):
        self.assertEqual(self.tool.inputs[0].inputBinding.prefix, '--input')

    def test_load_separate(self):
        self.assertTrue(self.tool.inputs[0].inputBinding.separate)

    def test_load_itemSeparator(self):
        self.assertEqual(self.tool.inputs[0].inputBinding.itemSeparator, ';')

    def test_load_valueFrom(self):
        self.assertEqual(self.tool.inputs[0].inputBinding.valueFrom, 'here')

    def test_load_shellQuote(self):
        self.assertTrue(self.tool.inputs[0].inputBinding.shellQuote)


class TestOutputsParser(TestImport):

    def test_init_input(self):
        self.assertEqual(self.tool.outputs[0].id, 'OUTPUT1')

    def test_load_label(self):
        self.assertEqual(self.tool.outputs[0].label, 'label_out')

    def test_load_secondaryFiles(self):
        self.assertEqual(self.tool.outputs[0].secondaryFile, 'sec_file_out')

    def test_load_format(self):
        self.assertEqual(self.tool.outputs[0].format, 'format_1930')

    def test_load_streamable(self):
        self.assertTrue(self.tool.outputs[0].streamable)

    def test_load_doc(self):
        self.assertEqual(self.tool.outputs[0].doc, 'documentation_out')

    def test_load_type(self):
        self.assertEqual(self.tool.outputs[0].type, 'File')


class TestOutputBindingParser(TestImport):

    def test_load_glob(self):
        self.assertEqual(self.tool.outputs[0].outputBinding.glob, 'find')

    def test_load_loadContents(self):
        self.assertTrue(self.tool.outputs[0].outputBinding.loadContents)

    def test_load_outputEval(self):
        self.assertEqual(self.tool.outputs[0].outputBinding.outputEval, "eval")



#!/usr/bin/env python

'''
Unit tests for cwlgen library
'''

#  Import  ------------------------------ 

# General libraries
import os
import filecmp
import unittest

# External libraries
import cwlgen


#  Class(es)  ------------------------------

class TestCommandLineTool(unittest.TestCase):

    def setUp(self):
        self.cwl = cwlgen.CommandLineTool(tool_id='an_id', label='a description ' + \
                                                                 'with spaces.', base_command='a_command')

        self.array_basecommand = cwlgen.CommandLineTool(tool_id="base-command-array", base_command=["base", "command"])

    def test_init(self):
        self.assertEqual(self.cwl.id, 'an_id')
        self.assertEqual(self.cwl.label, 'a description with spaces.')
        self.assertEqual(self.cwl.baseCommand, 'a_command')
        self.assertListEqual(self.cwl.inputs, [])
        self.assertListEqual(self.cwl.outputs, [])
        self.assertIsNone(self.cwl.doc)

    def test_array_basecommand(self):
        dict_test = self.array_basecommand.get_dict()
        self.assertIn('baseCommand', dict_test)

    def test_codes(self):
        code_tool = cwlgen.CommandLineTool()
        code_tool.permanentFailCodes.extend([400, 401, 500])
        code_tool.temporaryFailCodes.append(408)
        code_tool.successCodes = []  # empty (falsy)

        dict_test = code_tool.get_dict()
        self.assertIn('permanentFailCodes', dict_test)
        self.assertEqual(len(dict_test['permanentFailCodes']), 3)
        self.assertIn('temporaryFailCodes', dict_test)
        self.assertEqual(len(dict_test['temporaryFailCodes']), 1)
        self.assertNotIn('successCodes', dict_test)

    """
    def test_export(self):
        tmp_file = 'test_export.tmp'
        expected_file = os.path.dirname(__file__) + '/test_export.cwl'
        self.cwl.export(tmp_file)
        try:
            self.assertTrue(filecmp.cmp(expected_file, tmp_file))
        finally:
            os.remove(tmp_file)

    def test_full_export(self):
        tmp_file = 'test_full_export.tmp'
        expected_file = os.path.dirname(__file__) + '/test_full_export.cwl'
        self.cwl.doc = "documentation"
        an_input = cwlgen.CommandInputParameter('an_in_id', param_type='File')
        self.cwl.inputs.append(an_input)
        an_output = cwlgen.CommandOutputParameter('an_out_id', param_type='File')
        self.cwl.outputs.append(an_output)
        self.cwl.export(tmp_file)
        try:
            self.assertTrue(filecmp.cmp(expected_file, tmp_file))
        finally:
            os.remove(tmp_file)
    """


class TestWorkflow(unittest.TestCase):

    def setUp(self):
        self.workflow = cwlgen.Workflow("test-workflow", label="testing-workflow", doc="A method for testing "
                                                                                       "the features of a workflow.")
        self.workflow.inputs.append(cwlgen.InputParameter("inputParam1", "input-parameter"))
        self.workflow.steps.append(cwlgen.WorkflowStep("step1", run="mytool.cwl"))
        self.workflow.outputs.append(cwlgen.WorkflowOutputParameter("outputFile", output_source="step/output"))

    def test_init(self):
        self.assertEqual(self.workflow.id, "test-workflow")
        self.assertEqual(self.workflow.label, "testing-workflow")
        self.assertEqual(self.workflow.doc, "A method for testing the features of a workflow.")

    def test_added_inputs(self):
        self.assertEqual(len(self.workflow.inputs), 1)
        self.assertEqual(self.workflow.inputs[0].id, "inputParam1")

    def test_added_outputs(self):
        self.assertEqual(len(self.workflow.outputs), 1)
        self.assertEqual(self.workflow.outputs[0].id, "outputFile")

    def test_added_steps(self):
        self.assertEqual(len(self.workflow.steps), 1)
        step = self.workflow.steps[0]
        self.assertEqual(step.id, "step1")
        self.assertEqual(step.run, "mytool.cwl")

    def test_serialization(self):
        d = self.workflow.get_dict()
        self.assertIsInstance(d, dict)


class TestWorkflowStepInput(unittest.TestCase):
    def test_optional_source(self):
        wsi = cwlgen.WorkflowStepInput("test", None).get_dict()
        self.assertNotIn("source", wsi)

    def test_required_source(self):
        source = "step_id/source"
        wsi = cwlgen.WorkflowStepInput("required-input", source=source).get_dict()
        self.assertIn("source", wsi)
        self.assertEqual(wsi["source"], source)


class TestSubworkflow(unittest.TestCase):
    def setUp(self):
        self.subworkflow = cwlgen.Workflow("subworkflow", doc="This is the subworkflow")
        self.workflow = cwlgen.Workflow("workflow", doc="this will embed this subworkflow")

        self.workflow.steps.append(cwlgen.WorkflowStep("subworkflow-step", run=self.subworkflow))

    def test_subworkflow(self):
        sw = self.workflow.steps[0].run
        self.assertIsInstance(sw, cwlgen.Workflow)
        self.assertEqual(sw.id, self.subworkflow.id)

    def test_serialization(self):
        d = self.workflow.get_dict()
        self.assertIn("steps", d)
        steps = d["steps"]  # this should be a dictionary, so we look at 'key in dict'
        step_id = self.workflow.steps[0].id
        self.assertIn(step_id, steps)
        sw_step = steps[step_id]
        self.assertIsInstance(sw_step, dict)
        self.assertIn("run", sw_step)
        sw = sw_step["run"]
        self.assertIn("class", sw)
        sw_class = sw["class"]
        self.assertEqual(sw_class, "Workflow")


class TestParameter(unittest.TestCase):

    def setUp(self):
        self.param = cwlgen.Parameter('an_id', param_type='File', label='a_label', \
                                      doc='a_doc', param_format='a_format', \
                                      streamable=True, secondary_files='sec_files')
        self.nonfile_param = cwlgen.Parameter('non-file', param_type="string", label="a string",
                                              streamable=True, secondary_files=[".txt"], doc="documentation here")

        self.array_param = cwlgen.Parameter('an array', param_type='string[]', label="an array of strings")

    def test_init(self):
        self.assertEqual(self.param.id, 'an_id')
        self.assertEqual(self.param.type, 'File')
        self.assertEqual(self.param.doc, 'a_doc')
        self.assertEqual(self.param.format, 'a_format')
        self.assertEqual(self.param.label, 'a_label')
        self.assertEqual(self.param.secondaryFiles, 'sec_files')
        self.assertTrue(self.param.streamable)

    def test_get_dict(self):
        dict_test = self.param.get_dict()
        self.assertEqual(dict_test['type'], 'File')
        self.assertEqual(dict_test['doc'], 'a_doc')
        self.assertEqual(dict_test['format'], 'a_format')
        self.assertEqual(dict_test['label'], 'a_label')
        self.assertEqual(dict_test['secondaryFiles'], 'sec_files')
        self.assertTrue(dict_test['streamable'])

    # def test_nonfile_get_dict(self):
    #     dict_test = self.nonfile_param.get_dict()
    #     self.assertEqual(dict_test['type'], 'string')
    #     self.assertEqual(dict_test['doc'], self.nonfile_param.doc)
    #     self.assertNotIn('secondaryFiles', dict_test)
    #     self.assertNotIn('streamable', dict_test)
    #     self.assertNotIn('format', dict_test)

    def test_array(self):
        dict_test = self.array_param.get_dict()
        td = dict_test['type']
        self.assertIsInstance(td, dict)
        self.assertEqual(td['type'], 'array')
        self.assertEqual(td['items'], self.array_param.type.items)


class TestCommandInputParameter(unittest.TestCase):

    def setUp(self):
        self.frst_inp = cwlgen.CommandInputParameter('frst_id', param_type='File', \
                                                     label='a_label', \
                                                     default='def_value')
        binding = cwlgen.CommandLineBinding(position=2, prefix='--prefix')
        self.scnd_inp = cwlgen.CommandInputParameter('scnd_id', param_type='File', \
                                                     input_binding=binding)

    def test_init(self):
        # Test first input
        self.assertEqual(self.frst_inp.id, 'frst_id')
        self.assertEqual(self.frst_inp.type, 'File')
        self.assertEqual(self.frst_inp.label, 'a_label')
        self.assertEqual(self.frst_inp.default, 'def_value')
        # Test second input
        self.assertEqual(self.scnd_inp.id, 'scnd_id')
        self.assertEqual(self.scnd_inp.type, 'File')
        self.assertEqual(self.scnd_inp.inputBinding.position, 2)
        self.assertEqual(self.scnd_inp.inputBinding.prefix, '--prefix')

    def test_get_dict(self):
        # Test first input
        dict_frst = self.frst_inp.get_dict()
        self.assertEqual(dict_frst['type'], 'File')
        self.assertEqual(dict_frst['default'], 'def_value')
        self.assertEqual(dict_frst['label'], 'a_label')
        # Test second input
        dict_scnd = self.scnd_inp.get_dict()
        self.assertEqual(dict_scnd['type'], 'File')
        self.assertEqual(dict_scnd['inputBinding']['prefix'], '--prefix')
        self.assertEqual(dict_scnd['inputBinding']['position'], 2)

    def test_default(self):
        df = "DefaultParam"
        t = cwlgen.CommandInputParameter("has-default", default=df)
        d = t.get_dict()
        self.assertIn("default", d)
        self.assertEqual(t.default, d["default"])


class TestStep(unittest.TestCase):

    def test_ignore_id(self):
        step = cwlgen.WorkflowStep("identifier", "run")
        d = step.get_dict()
        self.assertNotIn("id", d)
        self.assertNotIn("ignore_attributes", d)

    def test_include_id(self):
        step = cwlgen.WorkflowStep("identifier", "run")
        step.ignore_attributes = None
        d = step.get_dict()
        self.assertIn("id", d)
        self.assertNotIn("ignore_attributes", d)


class TestCommandOutputParameter(unittest.TestCase):

    def setUp(self):
        self.outp = cwlgen.CommandOutputParameter('an_out_id', param_type='File')

    def test_init(self):
        self.assertEqual(self.outp.id, 'an_out_id')
        self.assertEqual(self.outp.type, 'File')

    def test_get_dict(self):
        dict_test = self.outp.get_dict()
        self.assertEqual(dict_test['type'], 'File')

    def test_empty_outputbinding(self):
        s = cwlgen.CommandOutputParameter("empty", output_binding=cwlgen.CommandOutputBinding()).get_dict()
        self.assertNotIn('outputBinding', s)
        self.assertEqual(s, {'id': 'empty'})

    def test_nonempty_outputbinding(self):
        s = cwlgen.CommandOutputParameter(
            "nonempty",
            output_binding=cwlgen.CommandOutputBinding(glob="$(inputs.test)")).get_dict()
        self.assertEqual(s, {'id': 'nonempty', 'outputBinding': {'glob': '$(inputs.test)'}})


class TestCommandLineBinding(unittest.TestCase):

    def setUp(self):
        self.line_binding = cwlgen.CommandLineBinding(load_contents=True, position=1, prefix='--prefix', separate=True,
                                                      item_separator='-', shell_quote=True, value_from='text.txt')
        self.false_line_binding = cwlgen.CommandLineBinding(load_contents=False, position=0, prefix='', separate=False,
                                                            item_separator='', shell_quote=False, value_from="")

    def test_init(self):
        self.assertTrue(self.line_binding.loadContents)
        self.assertEqual(self.line_binding.position, 1)
        self.assertEqual(self.line_binding.prefix, '--prefix')
        self.assertTrue(self.line_binding.separate)
        self.assertEqual(self.line_binding.itemSeparator, '-')
        self.assertTrue(self.line_binding.shellQuote)
        self.assertEqual(self.line_binding.valueFrom, 'text.txt')

    def test_get_dict(self):
        dict_test = self.line_binding.get_dict()
        self.assertEqual(dict_test['position'], 1)
        self.assertEqual(dict_test['prefix'], '--prefix')
        self.assertTrue(dict_test['separate'])
        self.assertEqual(dict_test['itemSeparator'], '-')
        self.assertTrue(dict_test['shellQuote'])
        self.assertEqual(dict_test['valueFrom'], 'text.txt')

    def test_false_dict(self):
        d = self.false_line_binding.get_dict()
        self.assertIn("loadContents", d)
        self.assertIn("position", d)
        self.assertIn("prefix", d)
        self.assertIn("separate", d)
        self.assertIn("itemSeparator", d)
        self.assertIn("shellQuote", d)
        self.assertIn("valueFrom", d)


class TestCommandOutputBinding(unittest.TestCase):

    def setUp(self):
        self.out_binding = cwlgen.CommandOutputBinding(glob='file.txt', load_contents=True, \
                                                       output_eval='eval')

    def test_init(self):
        self.assertEqual(self.out_binding.glob, 'file.txt')
        self.assertTrue(self.out_binding.loadContents)
        self.assertEqual(self.out_binding.outputEval, 'eval')

    def test_get_dict(self):
        dict_test = self.out_binding.get_dict()
        self.assertEqual(dict_test['glob'], 'file.txt')
        self.assertTrue(dict_test['loadContents'])
        self.assertEqual(dict_test['outputEval'], 'eval')

###########  Main  ###########

# if __name__ == "__main__":
#     unittest.main()



from setuptools import setup

__version__ = "v0.0.4"

setup(name="illusional.cwlgen",
      version=__version__,
      description='Fork of python-cwlgen | Generation of CWL programmatically. Avaible types: CommandLineTool, '
                  'Workflow and Requirements',
      long_description=open("./README.md").read(),
      long_description_content_type="text/markdown",
      author='Michael Franklin | Kenzo-Hugo Hillion and Herve Menager',
      author_email='michael.franklin@petermac.org',
      license='MIT',
      keywords=['cwl'],
      install_requires=['six', 'ruamel.yaml>=0.15'],
      packages=["cwlgen"],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Topic :: Scientific/Engineering',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Environment :: Console',
      ],
      )



import unittest

from cwlgen import get_type_dict
from cwlgen.common import parse_type, CwlTypes, CommandInputArraySchema
import logging


class TestParamTyping(unittest.TestCase):

    def test_types(self):
        for cwl_type in CwlTypes.NON_NULL_TYPES:
            self.assertEqual(parse_type(cwl_type), cwl_type)

    def test_incorrect_type(self):
        invalid_type = "invalid"
        should_be_def_type = parse_type(invalid_type)
        self.assertNotEqual(should_be_def_type, invalid_type)
        self.assertEqual(should_be_def_type, CwlTypes.DEF_TYPE)

    def test_optional_string(self):
        for cwl_type in CwlTypes.NON_NULL_TYPES:
            optional_type = cwl_type + "?"
            self.assertEqual(parse_type(optional_type), optional_type)

    def test_typed_array(self):
        array_string_type = "string[]"
        q = parse_type(array_string_type)
        self.assertIsInstance(q, CommandInputArraySchema)
        self.assertEqual(q.items, "string")

    def test_incorrectly_typed_array(self):
        array_string_type = "invalid[]"
        q = parse_type(array_string_type)
        self.assertIsInstance(q, CommandInputArraySchema)
        self.assertNotEqual(q.items, "invalid")
        self.assertEqual(q.items, CwlTypes.DEF_TYPE)

    def test_optionally_typed_array(self):
        array_string_type = "string?[]"
        q = parse_type(array_string_type)
        self.assertIsInstance(q, CommandInputArraySchema)
        self.assertEqual(q.items, "string?")

    def test_optional_typed_array(self):
        optional_array_string_type = "string[]?"
        q = parse_type(optional_array_string_type)
        self.assertIsInstance(q, list)
        self.assertEqual(len(q), 2)
        null_idx = q.index(CwlTypes.DEF_TYPE)
        array_type = q[1 - null_idx]
        self.assertIsInstance(array_type, CommandInputArraySchema)
        self.assertEqual(array_type.items, "string")

    def test_command_input_array_schema(self):
        ar = CommandInputArraySchema(items="string")
        self.assertIsInstance(ar, CommandInputArraySchema)
        self.assertEqual(parse_type(ar), ar)
        self.assertEqual(ar.items, "string")

    def test_optional_type_input_array_schema(self):
        ar = CommandInputArraySchema(items="string?")
        self.assertIsInstance(ar, CommandInputArraySchema)
        self.assertEqual(parse_type(ar), ar)
        self.assertEqual(ar.items, "string?")

    def test_array_of_array_of_strings(self):
        arstr = CommandInputArraySchema(items="string")
        ar = CommandInputArraySchema(items=arstr)
        self.assertEqual(ar.get_dict(), {'type': 'array', 'items': {'type': 'array', 'items': 'string'}})

    def test_parse_array_of_string_array(self):
        ar = CommandInputArraySchema(items="string[]")
        self.assertEqual(ar.get_dict(), {'type': 'array', 'items': {'type': 'array', 'items': 'string'}})

    def test_parse_list_of_items(self):
        t = ["string", "int"]
        types = parse_type(t)
        self.assertListEqual(t, types)

    def test_intentional_fail(self):
        try:
            parse_type("not_a_type", requires_type=True)
            self.fail("Failed to throw exception")
        except Exception as e:
            self.assertTrue(True)

    def test_intentional_fail_not_string(self):
        try:
            parse_type({}, requires_type=True)
            self.fail("Failed to throw exception")
        except:
            self.assertTrue(True)

    def test_get_type_dict(self):

        self.assertEqual("any_string_input", get_type_dict("any_string_input"))
        self.assertListEqual(["any", "list", "of", "strings"], get_type_dict(["any", "list", "of", "strings"]))
        self.assertDictEqual({"any": "dict"}, get_type_dict({"any": "dict"}))

        try:
            get_type_dict(self)
            self.assertTrue(False, "Failed to throw exception")
        except Exception as e:
            self.assertTrue(True)
